#ifndef GENHVBOX_H
#define GENHVBOX_H
typedef enum {
	HVBOX_WIDGET,              /* user widget with custom sizes */
	HVBOX_HBOX,                /* box that arranges children horizontally */
	HVBOX_VBOX,                /* box that arranges children vertically */
	HVBOX_ROOT
} hvbox_type_t;

typedef enum {
	HVBOX_MINIMAL  = 0,    /* default: keep size to the minimum possible */
	/* 1..100 */           /* try to take up this much space of parent */
	HVBOX_FILL     = 101   /* fill up the remaining space; at the end of calculating the value for the level, allocate free space evenly among DIR_FILL boxes */
} hvbox_span_t;

typedef enum {
	HVBOX_JUSTIFY,           /* default: space evenly distributed */
	HVBOX_CENTER,
	HVBOX_LEFT_TOP,
	HVBOX_RIGHT_BOTTOM
} hvbox_align_t;

typedef struct hvbox_s hvbox_t;

struct hvbox_s {
	/* specified by the user */
	hvbox_type_t type;
	int w_min, h_min;             /* minimal width/height determined by the content */
	int padding;                  /* leave this many pixels empty space within the border of box/widget, on each side */
	void *widget;
	hvbox_span_t span;         /* width or height span (depending on parent); hvbox_span_t or in percent */
	hvbox_align_t align;       /* alignment in case there's more room than needed */
	hvbox_span_t alt_span;     /* span in the altenrative direction (TODO) */

	unsigned char fr_l, fr_r;  /* frame left/right: hbox/vbox frame size, leaving less room for widgets */
	unsigned char fr_t, fr_b;  /* frame top/bottom: hbox/vbox frame size, leaving less room for widgets */
	unsigned char spacing;     /* hbox/vbox: minimum space between two child (not before the first or after the last) */

	unsigned int hidden:1;     /* override size to 0x0 and do not descend into children on update */

	/* calculated */
	unsigned int dirty:1;      /* hvbox calls set this 1 if x, y, w or h changed after an update */
	unsigned int hidden_calc:1;/* 1 if this widget or any of the parent widgets is hidden */
	int w, h;                  /* calculated width/height; own padding included so this is content_size + 2*padding */
	int x, y;                  /* top left corner coordinates; this is the top-left pixel of the actual content */
	int avail;                 /* available space left after packing up all children */
	int num_ch;                /* number of children */
	int w_min_calc, h_min_calc; /* boxes: calculated minimum width and height (inherited from children); own padding included */

	/* original/old values for dirty */
	int ow, oh, ox, oy;

	/* tree management */
	hvbox_t *next;                /* next sibling on this level */
	hvbox_t *ch_first, *ch_last;  /* first and last child under this node */
	hvbox_t *parent;
};

/* Create a new root box */
hvbox_t *hvbox_new_root(int w, int h, void *widget);

/* Build the tree; allocate a new widget or box under a parent and append it
   at the end of the sibling list or insert it in front of it */
hvbox_t *hvbox_append_widget(hvbox_t *parent, int w_min, int h_min, void *widget);
hvbox_t *hvbox_insert_widget(hvbox_t *parent, int w_min, int h_min, void *widget);
hvbox_t *hvbox_append_box(hvbox_t *parent, hvbox_type_t type);
hvbox_t *hvbox_insert_box(hvbox_t *parent, hvbox_type_t type);

/* Allocate a new box without connecting it into a tree (leaving it floating,
   for a later insertion) */
hvbox_t *hvbox_new_floating_box(int w_min, int h_min, hvbox_type_t type);

/* Append or insert a floating box in an existing tree */
hvbox_t *hvbox_append_existing_box(hvbox_t *parent, hvbox_t *box);
hvbox_t *hvbox_insert_existing_box(hvbox_t *parent, hvbox_t *box);
hvbox_t *hvbox_append_existing_box_at(hvbox_t *sibling, hvbox_t *box); /* insert box after sibling in sibling's parent */
hvbox_t *hvbox_insert_existing_box_at(hvbox_t *sibling, hvbox_t *box); /* insert box before sibling in sibling's parent */

/* Recalculate w;h and x;y of each node */
int hvbox_update(hvbox_t *root);

/* Destroy a subtree or the root */
void hvbox_destroy(hvbox_t *from);

/* Unlink a subtree from its parent (but do not free any memory) */
void hvbox_unlink(hvbox_t *from);

/* Widget accessors: at the end the caller is dealing with sizes and coordinates 
   of widgets; these macros calculate the true pixel coordinates of the
   top-left corner and bottom right corner, and the pixel width/height */
#define HVBOX_X(wg)  ((wg)->x)
#define HVBOX_Y(wg)  ((wg)->y)
#define HVBOX_W(wg)  ((wg)->w - 2*(wg)->padding)
#define HVBOX_H(wg)  ((wg)->h - 2*(wg)->padding)
#define HVBOX_X2(wg) (HVBOX_X(wg) + HVBOX_W(wg))
#define HVBOX_Y2(wg) (HVBOX_Y(wg) + HVBOX_H(wg))

/* Return the pointer of the innermost node that includes x and y */
hvbox_t *hvbox_by_coord(hvbox_t *root, int x, int y);


/* -- Forach-like iterator: take the next node in a BFS - start with root -- */
typedef struct {
	int used[2], alloced[2], cursor[2];
	hvbox_t **open[2];
	int level;
} hvbox_iter_t;

/* start iterating from node - returns node */
hvbox_t *hvbox_bfs_first(hvbox_iter_t *iterator, hvbox_t *node);

/* return the next node using the iterator; returns NULL when there's no more */
hvbox_t *hvbox_bfs_next(hvbox_iter_t *iterator);

/* free all memory claimed by the iterator */
void hvbox_bfs_finish(hvbox_iter_t *iterator);

#endif
